"""Module for the base distribution and the scipy distribution."""

from __future__ import annotations
from abc import ABC, abstractmethod
from copy import deepcopy
from typing import List, Iterable, Dict, Sequence, Union

import numpy as np
import polars as pl
import pandas as pd


class BaseDistribution(ABC):
    """Abstract base class to define a distribution.

    All distributions should be derived from this class, and the following
    methods need to be implemented: _fit, draw, to_dict.
    """

    aliases: List[str] = []
    is_unique = False
    var_type: str = "unknown"

    @classmethod
    def fit(cls, series: Union[Sequence, pl.Series], *args, **kwargs) -> BaseDistribution:
        """Fit the distribution to the series.

        Parameters
        ----------
        series: pandas.Series
            Data to fit the distribution to.

        Returns
        -------
        BaseDistribution:
            Fitted distribution.
        """
        pd_series = cls._to_series(series)
        if len(pd_series) == 0:
            return cls.default_distribution()
        distribution = cls._fit(pd_series, *args, **kwargs)
        return distribution

    @staticmethod
    def _to_series(values: Union[Sequence, pl.Series]):
        if isinstance(values, pl.Series):
            series = values.drop_nulls()
        elif isinstance(values, pd.Series):
            series = pl.Series(values).drop_nulls()  # pylint: disable=assignment-from-no-return
        else:
            series_array = np.array(values)
            series_array = series_array[~np.isnan(series_array)]
            series = pl.Series(series_array)
        return series

    @classmethod
    @abstractmethod
    def _fit(cls, values: pl.Series) -> BaseDistribution:
        """See fit method, but does not need to deal with NA's."""

    @abstractmethod
    def draw(self) -> object:
        """Draw a random element from the fitted distribution."""

    def draw_reset(self) -> None:
        """Reset the drawing of elements to start again."""

    def __str__(self) -> str:
        """Create a human readable string of the object."""
        return str(self.to_dict())

    @abstractmethod
    def to_dict(self) -> Dict:
        """Convert the distribution to a dictionary."""

    @classmethod
    def from_dict(cls, dist_dict: dict) -> BaseDistribution:
        """Create a distribution from a dictionary."""
        return cls(**dist_dict["parameters"])

    def information_criterion(self, values: Iterable) -> float:  # pylint: disable=unused-argument
        """Get the AIC value for a particular set of values.

        Parameters
        ----------
        values: array_like
            Values to determine the AIC value of.
        """
        return 0.0

    @classmethod
    def is_named(cls, name: str) -> bool:
        """Check whether the name matches the distribution.

        Parameters
        ----------
        name: str
            Name to match to the distribution.

        Returns
        -------
        bool:
            Whether the name matches.
        """
        return name in cls.aliases or name == type(cls).__name__ or name == cls.__name__

    @classmethod
    def fit_kwargs(cls, name: str) -> Dict:  # pylint: disable=unused-argument
        """Extra fitting arguments.

        Parameters
        ----------
        name: str
            Name to be matched.

        Returns
        -------
        dict:
            Keyword arguments extracted from the name.
        """
        return {}

    @property
    def name(self) -> str:
        """Return the name used in the metadata file."""
        return self.aliases[0]

    @classmethod
    @abstractmethod
    def default_distribution(cls) -> BaseDistribution:
        """Get a distribution with default parameters."""
        return cls()


class CategoricalDistribution(BaseDistribution):
    """Base Class for categorical distributions."""

    var_type = "categorical"


class DiscreteDistribution(BaseDistribution):
    """Base Class for discrete distributions."""

    var_type = "discrete"


class ContinuousDistribution(BaseDistribution):
    """Base Class for continuous distributions."""

    var_type = "continuous"


class StringDistribution(BaseDistribution):
    """Base Class for string distributions."""

    var_type = "string"


class DateTimeDistribution(BaseDistribution):
    """Base Class for date-time distributions."""

    var_type = "datetime"


class DateDistribution(BaseDistribution):
    """Base Class for date distributions."""

    var_type = "date"


class TimeDistribution(BaseDistribution):
    """Base Class for time distributions."""

    var_type = "time"


class ScipyDistribution(BaseDistribution):
    """Base class for numerical Scipy distributions.

    This base class makes it easy to implement new numerical
    distributions. One could also use this base class for non-scipy
    distributions, in which case the distribution class should implement
    logpdf, rvs and fit methods.
    """

    @property
    def n_par(self) -> int:
        """int: Number of parameters for distribution."""
        return len(self.par)

    def __getattr__(self, attr: str):
        """Get attribute for easy access to parameters.

        Parameters
        ----------
        attr:
            Attribute to retrieve. If the attribute is a parameter
            name, then retrieve that parameter, otherwise use the default
            implementation for getting the attribute.

        Returns
        -------
        object:
            Parameter or attribute.
        """
        if attr != "par" and attr in self.par:
            return self.par[attr]
        return object.__getattribute__(self, attr)

    @classmethod
    def _fit(cls, values):
        if len(values) == 0:
            return cls.default_distribution()
        param = cls.dist_class.fit(values)
        return cls(*param)

    def to_dict(self):
        return {
            "name": self.name,
            "parameters": deepcopy(self.par),
        }

    def draw(self):
        return self.dist.rvs()

    def information_criterion(self, values):
        vals = self._to_series(values)
        if len(vals) == 0:
            return 2*self.n_par
        return self._information_criterion(vals)

    def _information_criterion(self, values):
        return 2*self.n_par - 2*np.sum(self.dist.logpdf(values))



"""Distributions for variables.

These distributions can be fit to datasets/series so that the synthesis is
somewhat realistic. The concept of distributions here is not only for
numerical data, but also for generating strings for example.
"""  # pylint: disable=invalid-name

from metasynth.distribution.categorical import MultinoulliDistribution
from metasynth.distribution.continuous import NormalDistribution
from metasynth.distribution.continuous import UniformDistribution
from metasynth.distribution.discrete import DiscreteUniformDistribution
from metasynth.distribution.discrete import PoissonDistribution
from metasynth.distribution.regex import RegexDistribution
from metasynth.distribution.base import DiscreteDistribution
from metasynth.distribution.base import StringDistribution
from metasynth.distribution.base import ContinuousDistribution
from metasynth.distribution.base import CategoricalDistribution
from metasynth.distribution.faker import FakerDistribution



"""Module containing an interface to the faker package."""
from typing import Iterable

from faker import Faker

from metasynth.distribution.base import StringDistribution


class FakerDistribution(StringDistribution):
    """Distribution for the faker package.

    This is mainly an interface for the faker package, so that it
    can be used within the MetaSynth package. It doesn't have any
    true fitting/statistical inference method, so it has to be manually
    selected.

    Parameters
    ----------
    faker_type: str
        The provider function in the faker package, e.g. 'city' or 'ipv4', etc.
    locale: str
        Locale used for the faker package.
    """

    aliases = ["FakerDistribution", "faker"]

    def __init__(self, faker_type: str, locale: str="en_US"):
        self.faker_type: str = faker_type
        self.locale: str = locale
        self.fake: Faker = Faker(locale=locale)

    @classmethod
    def _fit(cls, values, faker_type: str="city", locale: str="en_US"):  # pylint: disable=arguments-differ
        """Select the appropriate faker function and locale."""
        return cls(faker_type, locale)

    def __str__(self):
        return f"faker.{self.faker_type}.{self.locale}"

    def to_dict(self):
        return {
            "name": self.name,
            "parameters": {
                "faker_type": self.faker_type,
                "locale": self.locale
            }
        }

    def draw(self):
        return getattr(self.fake, self.faker_type)()

    @classmethod
    def is_named(cls, name):
        if name == cls.__name__:
            return True
        return name.startswith("faker") and len(name.split(".")) >= 2

    @classmethod
    def fit_kwargs(cls, name):
        if name == cls.__name__:
            return {}
        split_name = name.split(".")
        if len(split_name) == 2:
            return {"faker_type": split_name[1]}
        return {"faker_type": split_name[1],
                "locale": split_name[2]}

    def information_criterion(self, values: Iterable) -> float:
        return 99999

    @classmethod
    def default_distribution(cls):
        return cls("city")



"""Module containing categorical distributions."""

from typing import Union

import pandas as pd
import numpy as np
import numpy.typing as npt
import polars as pl

from metasynth.distribution.base import CategoricalDistribution


class MultinoulliDistribution(CategoricalDistribution):
    """Categorical distribution that stores category labels and probabilities.

    Parameters
    ----------
    labels: list of str
        List containing the label belonging to each category.
    probs: list of int
        List containing the probability of each category.
        Probabilities will be normalized, so frequencies are valid too.
    """

    aliases = ["MultinoulliDistribution", "categorical"]

    def __init__(self, labels: npt.NDArray[np.str_],
                 probs: npt.NDArray[np.float_]):
        self.labels = labels
        self.probs = probs
        if np.sum(self.probs) != 1:
            self.probs = self.probs/np.sum(self.probs)

    @classmethod
    def _fit(cls, values: pl.Series):
        labels, counts = np.unique(values, return_counts=True)
        probs = counts/np.sum(counts)
        return cls(labels.astype(str), probs)

    @property
    def par_dict(self):
        """Get labels and probabilities as a dictionary."""
        return dict(zip(self.labels, self.probs))

    def to_dict(self):
        dist_dict = {}
        dist_dict["name"] = type(self).__name__
        dist_dict["parameters"] = {"labels": self.labels,
                                   "probs": self.probs}
        return dist_dict

    def __str__(self):
        return str(self.to_dict())

    def draw(self):
        return str(np.random.choice(self.labels, p=self.probs))

    def information_criterion(self,
                              values: Union[pd.Series, pl.Series, npt.NDArray[np.str_]]
                              ) -> float:
        values_array = np.array(values, dtype=str)
        labels, counts = np.unique(values_array, return_counts=True)
        log_lik = 0.0
        pdict = self.par_dict
        for lab, count in zip(labels, counts):
            # account for missing values / missing categories
            # by setting default of .get to 1 (add log(1)=0 to log_lik)
            log_lik += count * np.log(pdict.get(lab, 1))
        return 2*(len(self.probs) - 1) - 2 * log_lik

    @classmethod
    def default_distribution(cls):
        return cls(["a", "b", "c"], [10, 4, 20])



"""MetaSynth: a package for creating synthetic datasets.

One part concerns the creation of the statistical metadata from the
original data, while the other part creates a synthetic dataset from the
metadata.
"""

from metasynth.var import MetaVar
from metasynth.dataset import MetaDataset

from . import _version
__version__ = _version.get_versions()['version']



from pytest import mark, raises

from metasynth.distribution.categorical import MultinoulliDistribution
from metasynth.distribution.continuous import UniformDistribution,\
    NormalDistribution
from metasynth.distribution.faker import FakerDistribution
from metasynth.distribution.regex import RegexDistribution
from metasynth.distribution.discrete import DiscreteUniformDistribution
from metasynth.disttree import get_disttree


@mark.parametrize(
    "dist_str,dist",
    [
        ("categorical", MultinoulliDistribution),
        ("uniform", UniformDistribution),
        ("normal", NormalDistribution),
        ("gaussian", NormalDistribution),
        ("faker.city", FakerDistribution),
        ("faker.city.nl_NL", FakerDistribution),
        ("regex", RegexDistribution),
        ("discrete_uniform", DiscreteUniformDistribution)
    ]
)
def test_util(dist_str, dist):
    dist_tree = get_disttree()
    dist_class, _ = dist_tree.find_distribution(dist_str)
    assert dist == dist_class
    if dist_str.startswith("faker"):
        with raises(ValueError):
            dist_tree.find_distribution("this is not a distribution")
    new_class, _ = dist_tree.find_distribution(dist_class.__name__)
    assert new_class == dist_class



"""Conversion of pandas dataframes to MetaSynth datasets."""   # pylint: disable=invalid-name

from __future__ import annotations

from copy import deepcopy
from datetime import datetime
from importlib.resources import read_text
import json
import pathlib
from typing import Union, List, Dict, Any, Sequence, Optional

import numpy as np
import polars as pl
import jsonschema

from metasynth.var import MetaVar
from metasynth._version import get_versions
from metasynth.disttree import get_disttree


class MetaDataset():
    """MetaSynth dataset consisting of variables.

    The MetaSynth dataset structure that is most easily created from
    a pandas dataset with the from_dataframe class method.

    Parameters
    ----------
    meta_vars:
        List of variables representing columns in a dataframe.
    n_rows:
        Number of rows in the original dataframe.
    privacy_package:
        Package that supplies the distributions.
    """

    def __init__(self, meta_vars: List[MetaVar],
                 n_rows: Optional[int]=None,
                 privacy_package: Optional[str]=None):
        self.meta_vars = meta_vars
        self.n_rows = n_rows
        self.privacy_package = privacy_package

    @property
    def n_columns(self) -> int:
        """int: Number of columns of the original dataframe."""
        return len(self.meta_vars)

    @classmethod
    def from_dataframe(cls,
                       df: pl.DataFrame,
                       spec: Optional[dict[str, dict]] = None,
                       privacy_package: Optional[str]=None,
                       **privacy_kwargs):
        """Create dataset from a Pandas dataframe.

        The pandas dataframe should be formatted already with the correct
        datatypes.

        Parameters
        ----------
        df:
            Pandas dataframe with the correct column dtypes.
        spec:
            Column specifications to modify the defaults. For each of the columns additional
            directives can be supplied here. There are 3 different directives currently supported:

            distribution

            Set the distribution, either with a string that gives one of their
            aliases or an actually fitted BaseDistribution. For example:
            {"distribution": "NormalDistribution"} which is the same as
            {"distribution": NormalDistribution} or
            {"distribution": NormalDistribution(0, 1)}, which are all ways to set a variable
            to a normal distribution. Note that the first two do not set the parameters
            of the distribution, while the last does.

            unique

            To set a column to be unique/key.
            This is only available for the integer and string datatypes. Setting a variable
            to unique ensures that the synthetic values generated for this variable are unique.
            This is useful for ID or primary key variables, for example. The parameter...
            is ignored when the distribution is set manually. For example:
            {"unique": True}, which sets the variable to be unique or {"unique": False} which
            forces the variable to be not unique. If the uniqueness is not specified, it is
            assumed to be not unique, but might give a warning if it is detected that it might
            be.

            description

            Set the description of a variable: {"description": "Some description."}

            fit_kwargs

            Some distributions such as the FakerDistribution and RegexDistribution can take
            extra fitting arguments. For example for the RegexDistribution one can set the
            the speed of the computation by {"mode": "fast"} or {"mode": "slow"}. Be sure to set
            the distribution as well.

            Any number of the above directives may be set for any number of variables.

        privacy_package:
            Package that contains the implementations of the distributions.

        Returns
        -------
        MetaDataset:
            Initialized MetaSynth dataset.
        """
        distribution_tree = get_disttree(privacy_package, **privacy_kwargs)

        if spec is None:
            spec = {}
        else:
            spec = deepcopy(spec)

        all_vars = []
        for col_name in df.columns:
            series = df[col_name]
            col_spec = spec.get(col_name, {})
            dist = col_spec.pop("distribution", None)
            unq = col_spec.pop("unique", None)
            description = col_spec.pop("description", None)
            fit_kwargs = col_spec.pop("fit_kwargs", {})
            prop_missing = col_spec.pop("prop_missing", None)
            assert "fit_kwargs" not in col_spec
            if dist is None and len(fit_kwargs) > 0:
                raise ValueError(f"Got fit arguments for variable '{col_name}', but no "
                                 "distribution. Set the distribution manually to fix.")
            if len(col_spec) != 0:
                raise ValueError(f"Unknown spec items '{col_spec}' for variable '{col_name}'.")
            var = MetaVar.detect(series, description=description, prop_missing=prop_missing)
            if dist is None:
                var.fit(distribution_tree=distribution_tree, unique=unq, **fit_kwargs)
            else:
                var.fit(dist=dist, unique=unq, **fit_kwargs)

            all_vars.append(var)

        return cls(all_vars, len(df), privacy_package=privacy_package)

    def to_dict(self) -> Dict[str, Any]:
        """Create dictionary with the properties for recreation."""
        return {
            "n_rows": self.n_rows,
            "n_columns": self.n_columns,
            "provenance": {
                "created by": {
                    "name": "MetaSynth",
                    "version": get_versions()["version"],
                    "privacy": self.privacy_package,
                },
                "creation time": datetime.now().isoformat()
            },
            "vars": [var.to_dict() for var in self.meta_vars],
        }

    def __getitem__(self, key: Union[int, str]) -> MetaVar:
        """Return meta var either by variable name or index."""
        if isinstance(key, int):
            # If the key is an integer, give the ith variable.
            return self.meta_vars[key]
        if isinstance(key, str):
            # If the key is a string, return the first variable with that name.
            for var in self.meta_vars:
                if var.name == key:
                    return var
            raise KeyError(f"Cannot find variable '{key}'")
        raise TypeError(f"Cannot get item for key '{key}'")

    def __str__(self) -> str:
        """Create a readable string that shows the variables."""
        cur_str = "# Rows: "+str(self.n_rows)+"\n"
        cur_str += "# Columns: "+str(self.n_columns)+"\n"
        for var in self.meta_vars:
            cur_str += "\n"+str(var)+"\n"
        return cur_str

    @property
    def descriptions(self) -> dict[str, str]:
        """Return the descriptions of the columns."""
        return {var.name: var.description for var in self.meta_vars
                if var.name is not None and var.description is not None}

    @descriptions.setter
    def descriptions(self, new_descriptions: Union[dict[str, str], Sequence[str]]):
        if isinstance(new_descriptions, dict):
            for var_name, new_desc in new_descriptions.items():
                self[var_name].description = new_desc
        else:
            assert len(new_descriptions) == self.n_columns, (
                "Descriptions need to be either a dict or a "
                "sequence with the length of the number of variables.")
            for i_desc, new_desc in enumerate(new_descriptions):
                self[i_desc].description = new_desc

    def to_json(self, fp: Union[pathlib.Path, str], validate: bool=True) -> None:
        """Write the MetaSynth dataset to a JSON file.

        Optional validation against a JSON schema included in the package.

        Parameters
        ----------
        fp:
            File to write the dataset to.
        validate:
            Validate the JSON file with a schema.
        """
        self_dict = _jsonify(self.to_dict())
        if validate:
            schema = json.loads(read_text("metasynth.schema", "generative_metadata_format.json"))
            jsonschema.validate(instance=self_dict, schema=schema)
        with open(fp, "w", encoding="utf-8") as f:
            json.dump(self_dict, f, indent=4)

    @classmethod
    def from_json(cls, fp: Union[pathlib.Path, str], validate: bool=True) -> MetaDataset:
        """Read a MetaSynth dataset from a JSON file.

        Parameters
        ----------
        fp:
            Path to read the data from.
        validate:
            Validate the JSON file with a schema.

        Returns
        -------
        MetaDataset:
            A restored metadataset from the file.
        """
        with open(fp, "r", encoding="utf-8") as f:
            self_dict = json.load(f)

        if validate:
            schema = json.loads(read_text("metasynth.schema", "generative_metadata_format.json"))
            jsonschema.validate(instance=self_dict, schema=schema)

        n_rows = self_dict["n_rows"]
        meta_vars = [MetaVar.from_dict(d) for d in self_dict["vars"]]
        return cls(meta_vars, n_rows)

    def synthesize(self, n: int) -> pl.DataFrame:
        """Create a synthetic pandas dataframe.

        Parameters
        ----------
        n:
            Number of rows to generate.

        Returns
        -------
        pandas.DataFrame:
            Dataframe with the synthetic data.
        """
        synth_dict = {var.name: var.draw_series(n) for var in self.meta_vars}
        return pl.DataFrame(synth_dict)


def _jsonify(data):
    if isinstance(data, (list, tuple)):
        return [_jsonify(d) for d in data]
    if isinstance(data, dict):
        return {key: _jsonify(value) for key, value in data.items()}

    if isinstance(data, (np.int8, np.int16, np.int32, np.int64)):
        return int(data)
    if isinstance(data, np.ndarray):
        return _jsonify(data.tolist())
    return data



from random import random
from pathlib import Path

import pytest
import pandas as pd
import polars as pl
from pytest import mark

from metasynth.dataset import MetaDataset
from metasynth.var import MetaVar
from metasynth.disttree import get_disttree


dtypes = {
    "PassengerId": "int",
    "Survived": "category",
    "Pclass": "category",
    "Name": "string",
    "Sex": "category",
    "SibSp": "category",
    "Parch": "category",
    "Ticket": "string",
    "Cabin": "string",
    "Embarked": "category",
    "Age": "float",
    "Fare": "float",
}


def _read_csv(fp, dataframe_lib):
    if dataframe_lib == "pandas":
        df = pd.read_csv(fp, dtype=dtypes)
        return df.iloc[:100]
    else:
        df = pl.read_csv(fp, dtypes={x: pl.Categorical for x, x_type in dtypes.items() if x_type == "category"})
        return df[:100]


@mark.parametrize("dataframe_lib", ["polars", "pandas"])
def test_dataset(tmp_path, dataframe_lib):
    titanic_fp = Path("tests", "data", "titanic.csv")
    tmp_fp = tmp_path / "tmp.json"
    df = _read_csv(titanic_fp, dataframe_lib)
    dataset = MetaDataset.from_dataframe(
        df,
        spec={
                "Name": {"prop_missing": 0.5},
                "Ticket": {"description": "test_description"},
                "Fare": {"distribution": "normal"},
                "PassengerId": {"unique": True},
             })

    def check_dataset(dataset):
        assert dataset.n_columns == 12
        assert dataset.n_rows == 100
        sub_types = {
            "int": "discrete",
            "string": "string",
            "category": "categorical",
            "float": "continuous",
        }
        for col, type_name in dtypes.items():
            meta_var_type = sub_types[type_name]
            assert dataset[col].var_type == meta_var_type
        with pytest.raises(KeyError):
            dataset["unknown"]
        with pytest.raises(IndexError):
            dataset[20]
        with pytest.raises(TypeError):
            dataset[MetaVar]

        assert "Rows" in str(dataset)

    check_dataset(dataset)
    dataset.to_json(tmp_fp)
    dataset = MetaDataset.from_json(tmp_fp)
    check_dataset(dataset)

    dataset.descriptions = {"Embarked": "Some description", "Sex": "Other description"}
    assert dataset.descriptions["Embarked"] == "Some description"
    with pytest.raises(AssertionError):
        dataset.descriptions = "12345"
    dataset.descriptions = {"Embarked": "New description"}
    assert dataset.descriptions["Embarked"] == "New description"
    assert dataset.descriptions["Sex"] == "Other description"
    if isinstance(df, pl.DataFrame):
        col_names = df.columns
    else:
        col_names = list(df)
    dataset.descriptions = col_names
    for name in col_names:
        print(name, dataset.descriptions[name])
        assert dataset.descriptions[name] == name


def test_distributions(tmp_path):
    tmp_fp = tmp_path / "tmp.json"

    dist_tree = get_disttree()
    for var_type in dist_tree.all_var_types:
        for dist in dist_tree.get_dist_list(var_type):
            var = MetaVar(var_type, name="None", distribution=dist.default_distribution(),
                          prop_missing=random())
            dataset = MetaDataset([var], n_rows=10)
            dataset.to_json(tmp_fp)



# based on https://github.com/pypa/sampleproject - MIT License
from pathlib import Path
from setuptools import setup, find_packages

import versioneer


readme_path = Path(__file__).parent.absolute() / "README.md"
with open(readme_path, "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name='metasynth',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author='SODA Development Team',
    description='Package for creating synthetic datasets from datasets',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(exclude=['data', 'docs', 'tests', 'examples']),
    package_data={"metasynth": ["py.typed"]},
    include_package_data=True,
    python_requires='~=3.7',
    install_requires=[
        "pandas",
        "polars>=0.14.17",
        "pyarrow",  # Dependency of polars since we're converting from pandas.
        "scipy",
        "numpy>=1.20",
        "faker",
        "scikit-learn",
        "xmltodict",
        "jsonschema",
        "importlib-resources;python_version<'3.9'",
        "wget",
    ],
    entry_points={
        'metasynth.disttree': [
            "builtin = metasynth.disttree:BuiltinDistributionTree",
        ]
    }
)



import pandas as pd
import polars as pl
from pytest import mark

from metasynth.distribution.faker import FakerDistribution


@mark.parametrize("series_type", [pd.Series, pl.Series])
def test_faker(series_type):
    var = FakerDistribution.fit(series_type([1, 2, 3]))
    assert isinstance(var.to_dict(), dict)
    assert isinstance(var.draw(), str)
    assert var.is_named("faker.city.nl_NL")
    fit_kwargs = var.fit_kwargs("faker.city.nl_NL")
    assert fit_kwargs["faker_type"] == "city"
    assert fit_kwargs["locale"] == "nl_NL"
    fit_kwargs = var.fit_kwargs("faker.city")
    assert fit_kwargs["faker_type"] == "city"
    assert "locale" not in fit_kwargs
    assert str(var).startswith("faker")



import string
from random import choice

import numpy as np
import pandas as pd
import polars as pl
from pytest import mark, raises

from metasynth.distribution.regex import RegexDistribution, UniqueRegexDistribution
from metasynth.distribution.regex.element import SingleRegex,\
    DigitRegex, AlphaNumericRegex, LettersRegex, LowercaseRegex, UppercaseRegex,\
    AnyRegex


@mark.parametrize("series_type", [pd.Series, pl.Series])
def test_regex_single_digit(series_type):
    series = series_type(["R123", "R837", "R354", "R456", "R578", "R699"])
    dist = RegexDistribution.fit(series)
    dist_unique = UniqueRegexDistribution.fit(series)
    assert dist.information_criterion(series) < dist_unique.information_criterion(series)

    def check_regex_dist(dist):
        assert len(dist.re_list) == 2
        assert isinstance(dist.re_list[0], SingleRegex)
        assert isinstance(dist.re_list[1], DigitRegex)
        assert dist.re_list[0].character_selection == ["R"]
        assert dist.re_list[1].min_digit == 3
        assert dist.re_list[1].max_digit == 3
        assert str(dist) == r"[R]\d{3,3}"

        for draw_str in [dist.draw() for _ in range(10)]:
            assert len(draw_str) == 4
            assert draw_str[0] == "R"

    check_regex_dist(dist)

    re_string_list = dist.to_dict()["parameters"]["re_list"]
    new_dist = RegexDistribution(re_string_list)
    check_regex_dist(new_dist)


@mark.parametrize("series_type", [pd.Series, pl.Series])
def test_regex_unique(series_type):
    series = series_type(["R1", "R2", "R3", "R4", "R5", "R6"])
    dist = UniqueRegexDistribution.fit(series)
    values = [dist.draw() for _ in range(10)]
    assert len(set(values)) == 10
    assert set(values) == set(["R" + x for x in string.digits])
    with raises(ValueError):
        dist.draw()
    dist.draw_reset()
    dist.draw()


@mark.parametrize(
    "dist_class,regex_str,regex_str_alt",
    [
        (AlphaNumericRegex, r"\w{1,1}", r"\w"),
        (LettersRegex, r"[a-zA-Z]{1,1}",  r"[a-zA-Z]"),
        (LowercaseRegex, r"[a-z]{1,1}", r"[a-z]"),
        (UppercaseRegex, r"[A-Z]{1,1}", r"[A-Z]"),
        (DigitRegex, r"\d{1,1}", r"\d"),
        (AnyRegex, r".[]{1,1}", r".[]"),
    ]
)
def test_optional_length(dist_class, regex_str, regex_str_alt):
    dist, _ = dist_class.from_string(regex_str)
    dist_alt, _ = dist_class.from_string(regex_str_alt)
    assert str(dist) == str(dist_alt)


@mark.parametrize(
    "digit_set,dist_class,regex_str,n_digits",
    [
        (string.ascii_letters+string.digits, AlphaNumericRegex, r"\w{10,10}", 10),
        (string.ascii_letters, LettersRegex, r"[a-zA-Z]{10,10}", 10),
        (string.digits, DigitRegex, r"\d{10,10}", 10),
        (string.ascii_lowercase, LowercaseRegex, r"[a-z]{10,10}", 10),
        (string.ascii_uppercase, UppercaseRegex, r"[A-Z]{10,10}", 10),
    ]
)
@mark.parametrize("series_type", [pd.Series, pl.Series])
def test_digits(digit_set, dist_class, regex_str, n_digits, series_type):
    def draw():
        draw_str = ""
        for _ in range(n_digits):
            draw_str += choice(digit_set)
        return draw_str

    series = series_type([draw() for _ in range(100)])
    dist = RegexDistribution.fit(series)
    assert len(dist.re_list) == 1
    assert isinstance(dist.re_list[0], dist_class)
    assert dist.re_list[0].min_digit == n_digits
    assert dist.re_list[0].max_digit == n_digits
    assert dist.to_dict()["parameters"]["re_list"][0][0] == regex_str
    assert np.all([len(dist.draw()) == n_digits for _ in range(100)])
    assert np.all([c in digit_set for c in dist.draw()])
    new_dist = dist_class.from_string(str(dist), 1.0)[0]
    with raises(ValueError):
        dist_class.from_string(str(dist), 3)[0]
    assert isinstance(new_dist, dist_class)
    assert new_dist.min_digit == dist.re_list[0].min_digit
    assert new_dist.max_digit == dist.re_list[0].max_digit





"""Module for distribution trees.

These are used to find/fit distributions that are available. See setup.py on how the
builtin distribution tree is registered.
"""

from __future__ import annotations

from abc import abstractmethod
from typing import List, Union
from typing import Type, Any, Optional
import warnings
import inspect
import pkg_resources

import polars as pl
import numpy as np

from metasynth.distribution.base import BaseDistribution
from metasynth.distribution.discrete import DiscreteUniformDistribution,\
    PoissonDistribution, UniqueKeyDistribution
from metasynth.distribution.continuous import UniformDistribution,\
    NormalDistribution, LogNormalDistribution, TruncatedNormalDistribution,\
    ExponentialDistribution
from metasynth.distribution.categorical import MultinoulliDistribution
from metasynth.distribution.regex.base import RegexDistribution,\
    UniqueRegexDistribution
from metasynth.distribution.faker import FakerDistribution
from metasynth.distribution.datetime import UniformDateDistribution,\
    UniformTimeDistribution, UniformDateTimeDistribution


class BaseDistributionTree():
    """Class that encapsulates a set of distributions.

    It has a property {var_type}_distributions for every var_type.
    """

    def __init__(self, **kwargs):
        # Perform internal consistency check.
        for var_type in self.all_var_types:
            for dist in self.get_dist_list(var_type):
                assert dist.var_type == var_type, (f"Error: Distribution tree is inconsistent for "
                                                   f"{dist}.")
        self.privacy_kwargs = kwargs

    @property
    @abstractmethod
    def discrete_distributions(self) -> List[Type[BaseDistribution]]:
        """Get the integer distributions."""

    @property
    @abstractmethod
    def continuous_distributions(self) -> List[Type[BaseDistribution]]:
        """Get continuous distributions."""

    @property
    @abstractmethod
    def categorical_distributions(self) -> List[Type[BaseDistribution]]:
        """Get categorical distributions."""

    @property
    @abstractmethod
    def string_distributions(self) -> List[Type[BaseDistribution]]:
        """Get categorical distributions."""

    @property
    @abstractmethod
    def date_distributions(self) -> List[Type[BaseDistribution]]:
        """Get categorical distributions."""

    @property
    @abstractmethod
    def time_distributions(self) -> List[Type[BaseDistribution]]:
        """Get categorical distributions."""

    @property
    @abstractmethod
    def datetime_distributions(self) -> List[Type[BaseDistribution]]:
        """Get categorical distributions."""

    def get_dist_list(self, var_type: str) -> List[Type[BaseDistribution]]:
        """Get all distributions for a certain variable type.

        Parameters
        ----------
        var_type:
            Variable type to get the distributions for.

        Returns
        -------
        list[Type[BaseDistribution]]:
            List of distributions with that variable type.
        """
        prop_str = var_type + "_distributions"
        if not hasattr(self, prop_str):
            raise ValueError(f"Unknown variable type '{var_type}' detected.")
        return getattr(self, prop_str)

    def fit(self, series: pl.Series, var_type: str,
            unique: Optional[bool]=False) -> BaseDistribution:
        """Fit a distribution to a series.

        Search for the distirbution within all available distributions in the tree.

        Parameters
        ----------
        series:
            Series to fit a distribution to.
        var_type:
            Variable type of the series.
        unique:
            Whether the variable should be unique or not.

        Returns
        -------
        BaseDistribution:
            Distribution fitted to the series.
        """
        dist_list = self.get_dist_list(var_type)
        if len(dist_list) == 0:
            raise ValueError(f"No available distributions with variable type: '{var_type}'")
        dist_instances = [d.fit(series, **self.privacy_kwargs) for d in dist_list]
        dist_aic = [d.information_criterion(series) for d in dist_instances]
        i_best_dist = np.argmin(dist_aic)
        warnings.simplefilter("always")
        if dist_instances[i_best_dist].is_unique and unique is None:
            warnings.warn(f"\nVariable {series.name} seems unique, but not set to be unique.\n"
                          "Set the variable to be either unique or not unique to remove this "
                          "warning.\n")
        if unique is None:
            unique = False

        dist_aic = [dist_aic[i] for i in range(len(dist_aic))
                    if dist_instances[i].is_unique == unique]
        dist_instances = [d for d in dist_instances if d.is_unique == unique]
        if len(dist_instances) == 0:
            raise ValueError(f"No available distributions for variable '{series.name}'"
                             f" with variable type '{var_type}' "
                             f"that have unique == {unique}.")
        return dist_instances[np.argmin(dist_aic)]

    @property
    def all_var_types(self) -> List[str]:
        """Return list of available variable types."""
        return [p[:-14] for p in dir(self.__class__)
                if isinstance(getattr(self.__class__, p), property) and p.endswith("_distributions")
                ]

    def find_distribution(self, dist_name: str) -> tuple[Type[BaseDistribution], dict[str, Any]]:
        """Find a distribution and fit keyword arguments from a name.

        This allows us to use 'faker.city' to generate a faker instance that generates cities.

        Parameters
        ----------
        dist_name:
            Name of the distribution, such as faker.city, DiscreteUniformDistribution or normal.

        Returns
        -------
        tuple[Type[BaseDistribution], dict[str, Any]]:
            A distribution and the arguments to create an instance.
        """
        for var_type in self.all_var_types:
            for dist_class in self.get_dist_list(var_type):
                if dist_class.is_named(dist_name):
                    return dist_class, dist_class.fit_kwargs(dist_name)
        raise ValueError(f"Cannot find distribution with name '{dist_name}'.")

    def fit_distribution(self, dist: Union[str, Type[BaseDistribution], BaseDistribution],
                         series: pl.Series, **fit_kwargs) -> BaseDistribution:
        """Fit a specific distribution to a series.

        In contrast the fit method, this needs a supplied distribution(type).

        Parameters
        ----------
        dist:
            Distribution to fit (if it is not already fitted).
        series:
            Series to fit the distribution to
        fit_kwargs:
            Extra fitting parameters that are specific to the distribution.

        Returns
        -------
        BaseDistribution:
            Fitted distribution.
        """
        dist_instance = None
        fit_kwargs.update(self.privacy_kwargs)

        if isinstance(dist, str):
            dist_class, new_fit_kwargs = self.find_distribution(dist)
            fit_kwargs.update(new_fit_kwargs)
            dist_instance = dist_class.fit(series, **fit_kwargs)
        elif inspect.isclass(dist) and issubclass(dist, BaseDistribution):
            dist_instance = dist.fit(series, **fit_kwargs)
        if isinstance(dist, BaseDistribution):
            dist_instance = dist

        if dist_instance is None:
            raise TypeError(
                f"Distribution with type {type(dist)} is not a BaseDistribution")

        return dist_instance

    def from_dict(self, var_dict: dict[str, Any]) -> BaseDistribution:
        """Create a distribution from a dictionary.

        Parameters
        ----------
        var_dict:
            Variable dictionary that includes the distribution properties.

        Returns
        -------
        BaseDistribution:
            Distribution representing the dictionary.
        """
        for dist_class in self.get_dist_list(var_dict["type"]):
            if dist_class.is_named(var_dict["distribution"]["name"]):
                return dist_class.from_dict(var_dict["distribution"])
        raise ValueError(f"Cannot find distribution with name '{var_dict['distribution']['name']}'"
                         f"and type '{var_dict['type']}'.")


class BuiltinDistributionTree(BaseDistributionTree):
    """Distribution tree that includes the builtin distributions."""

    @property
    def discrete_distributions(self) -> List[type]:
        return [DiscreteUniformDistribution, PoissonDistribution, UniqueKeyDistribution]

    @property
    def continuous_distributions(self) -> List[type]:
        return [UniformDistribution, NormalDistribution, LogNormalDistribution,
                TruncatedNormalDistribution, ExponentialDistribution]

    @property
    def categorical_distributions(self) -> List[type]:
        return [MultinoulliDistribution]

    @property
    def string_distributions(self) -> List[type]:
        return [RegexDistribution, UniqueRegexDistribution, FakerDistribution]

    @property
    def date_distributions(self) -> List[type]:
        return [UniformDateDistribution]

    @property
    def time_distributions(self) -> List[type]:
        return [UniformTimeDistribution]

    @property
    def datetime_distributions(self) -> List[type]:
        return [UniformDateTimeDistribution]


def get_disttree(target: Optional[Union[str, type, BaseDistributionTree]]=None, **kwargs
                 ) -> BaseDistributionTree:
    """Get a distribution tree.

    Parameters
    ----------
    target:
        Directive to get the distribution tree.

    Returns
    -------
    BaseDistributionTree:
        Distribution tree.
    """
    if target is None:
        target = "builtin"
    if isinstance(target, BaseDistributionTree):
        return target
    if isinstance(target, type):
        return target()

    all_disttrees = {
        entry.name: entry
        for entry in pkg_resources.iter_entry_points("metasynth.disttree")
    }
    try:
        return all_disttrees[target].load()(**kwargs)
    except KeyError as exc:
        raise ValueError(f"Cannot find distribution tree with name '{target}'.") from exc
